import 'token.dart';
import 'unprocessed_token.dart';

abstract class Lexer {
  Lexer({
    this.stripnl = true,
    this.stripall = false,
    this.ensurenl = true,
    this.tabsize = 0,
    this.encoding = 'guess',
    this.debuggable = false,
  });
  final bool stripnl;
  final bool stripall;
  final bool ensurenl;
  final int tabsize;
  final String encoding;
  final bool debuggable;

  String get root;
  String? get name => null;
  List<String> get aliases => [];
  List<String> get filenames => [];
  List<String> get aliasFilenames => [];
  List<String> get mimetypes => [];
  int get priority => 0;

  // Return an iterable of (index, tokentype, value) pairs where "index"
  // is the starting position of the token within the input text.
  //
  // In subclasses, implement this method as a generator to
  // maximize effectiveness.

  Iterable<UnprocessedToken> getTokensUnprocessed(String text);

  String processedString(
      {Iterable<UnprocessedToken>? unprocessedTokens, String? text}) {
    unprocessedTokens ??= getTokensUnprocessed(text!);
    var out = '';
    String? currentTokenName;
    var newlineNeeded = false;
    for (var token in unprocessedTokens) {
      if (token.token == Token.Text) {
        if (currentTokenName != null) {
          out += ')';
          currentTokenName = null;
        }
        out += token.match;
      } else {
        if (token.tokenName != currentTokenName) {
          if (currentTokenName != null) {
            out += ')';
            if (newlineNeeded == true) {
              out += '\n';
              newlineNeeded = false;
            }
          }
          out += token.tokenName + '(';
          currentTokenName = token.tokenName;
          if (token.match.contains('\n')) {
            newlineNeeded = true;
          }
        }
        out += token.prettyMatch;
      }
    }
    if (currentTokenName != null) {
      out += ')';
      if (newlineNeeded == true) {
        out += '\n';
      }
    }
    return out;
  }
}
